/*

             Copyright 1994,1995, 2000 Digital Equipment Corporation
	     Copyright 2000 Compaq Computer Corporation
                        All rights reserved.

   Redistribution and use in source and binary forms are
   permitted provided that: 1) source distributions retain this
   entire copyright notice and 2) distributions including
   binaries display the following acknowledgment in its
   documentation:

           "This product includes software developed by the
            Digital Equipment Corporation and Compaq Computer
	    Corporation"
*/

/*
 * All system-dependent information is in this file.
 * In addition, all data gathering and display information is
 * defined in this file. No changes to any other files should
 * be necessary to add new information to the application.
 */
/*
 * MKN: 01-Dec-2000
 * Remove UBC_status bar. Move clean/dirty into UBC_page display.
 *
 * MKN: 13-Nov-2000
 * New source for UBC cache hit/miss data
 * 
 * MKN: 23-Oct-2000
 * Add "syscall" bar
 *
 * MKN: 19-Oct-2000, Martin Knoblauch
 *
 * Bring back some V4 functionality to V5
 * Split active pages (V4) into active (V5) plus UBC pages
 * Add Intr bar
 * Try to cleanup some unnecessary code
 */
/* 25-May-2000 Bernard Laforgue
 * ported to UNIX V5.x
 * this file compiles for all versions depending on the variable DUVER
 * DUVER should be set with the last 3 digit of the OSFBASE subset
 * for example 505 for Tru64 UNIX V5.0A
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/table.h>
#include <sys/sysinfo.h>
#include <machine/hal_sysinfo.h>
#include <mach.h>
#include <nlist.h>
#include <paths.h>
#include <unistd.h>
#include <sys/vm.h>
#include <sys/file.h>

#include <sys/utsname.h>

#include "vmubc.h"

/**
*** MKN: Stuff for querying sysconfigdb
**/
#if DUVER >= 500
#include <cfg.h>
cfg_attr_t ubc_attr[4];
cfg_status_t cfg_retval;
long ubc_minpercent,ubc_maxpercent;
#endif

/*
** MKN: Only GETAPAGE and BIOSTATS necessary for V500
*/
struct nlist nl[] = {
    { "bio_stats" },
#define N_BIOSTATS 0
#if DUVER < 500
    { "ufs_getapage_stats" },
#define N_UFS_GETAPAGE_STATS 1
    { "ubc_pages" },
#define N_UBC_PAGES 2
    { "ubc_minpages" },
#define N_UBC_MINPAGES 3
    { "ubc_maxpages" },
#define N_UBC_MAXPAGES 4
    { "ubc_dirtypages" },
#define N_UBC_DIRTYPAGES 5
#endif DUVER < 500
    { NULL },
};
vm_size_t vm_swap_space, vm_total_swap_space;
vm_size_t ubc_minpages, ubc_maxpages, vm_managed_pages;
vm_size_t ubc_dirtypages, ubc_pages;
	

struct ufs_getapage_stats {
	long read_looks;
	long read_hits;
	long read_miss;
} pg_stats,old_pg_stats;

struct bio_stats {
	int	getblk_hits;
	int	getblk_misses;
	int	getblk_research;
	int	getblk_dupbuf;
	int	getnewbuf_calls;
	int	getnewbuf_buflocked;
	int	vflushbuf_lockskips;
	int	mntflushbuf_misses;
	int	mntinvalbuf_misses;
	int	vinvalbuf_misses;
	int	allocbuf_buflocked;
	int	ufssync_misses;
} bstats,old_bstats;

extern int firstrun; // MKN: Used for first-pass treatment
extern int interval; // MKN: Used to scale "rates" to second intervalls

static int kmem;
static int ncpus;

static struct tbl_sysinfo	 info;		/* current # ticks */
static struct tbl_sysinfo	 prev_info;	/* previous # ticks */
struct tick_struct	 *cpu_info;	/* per cpu current # ticks */
static struct tick_struct	 *prev_cpu_info;/* per cpu previous # ticks */
#if DUVER < 500
static struct vm_statistics	 vmstat;	/* virtual memory stats */
static struct vm_statistics	 prev_vmstat;	/* previous virtual memory stats */
#else /* DUVER < 500 */
static struct tbl_vmstats	 vmstat;	/* virtual memory stats */
static struct tbl_vmstats	 prev_vmstat;	/* previous virtual memory stats */
#endif /* DUVER < 500 */
static struct tbl_intr		 intrtb;	/* interrupt stats. MKN */
static struct tbl_intr		 prev_intrtb;	/* previous interrupt stats. MKN */
static struct tbl_scalls	 scalls;	/* syscall stats. MKN */
static struct tbl_scalls	 prev_scalls;	/* previous syscall stats. MKN */

/* forward function definitions */

extern int table ();  /* system call */
extern void draw_bar (int bar_num, int *states, int num_states);

void update_cpu_bar();
void update_memory_bar();
void update_UBC_usage_bar();
void update_UBC_cache_bar();
void update_metadata_bar();
void update_pagefault_bar();
void update_swap_bar();
void update_intr_bar(); // MKN.
void update_syscall_bar(); // MKN.
void init_cpu_bar();
void bad_func();

static BarInfo bar_info[] =
{
	{
		PROPORTION,		/* bar_type */
		1,			/* num_scales */
		update_cpu_bar,		/* update() */
		init_cpu_bar,		/* init() */
		"CPU\n(percent)",	/* label */
					/* legend */
/*
** MKN: Swap wait and idle
*/
	 	{5, {"User", "Nice", "System", "Wait","Idle" } },
	},
	{
	   	PROPORTION,		/* bar_type */
		1,			/* num_scales */
		update_memory_bar,	/* update() */
		NULL,			/* init() */
		"Memory\nPages",	/* label */
					/* legend */
/*
** MKN: Add UBC
*/
	 	{5, {"Active", "Inactive", "Wired", "UBC", "Free" } },
	},
	{
		PROPORTION_LIMIT,	/* bar_type */
		1,			/* num_scales */
		update_UBC_usage_bar,	/* update() */
		NULL,			/* init() */
      		"UBC Page\nUsage",	/* label */
					/* legend */
	 	{5, {"UBC-Min", "Clean", "Dirty", "VM", "VM-Only"} },
	},
	{
		LENGTH,		/* bar_type */
		2,			/* num_scales */
		update_UBC_cache_bar,	/* update() */
		NULL,			/* init() */
      		"UBC\nCache",	/* label */
					/* legend */
		{2, {"Hits", "Misses" } },
	},
	{
		LENGTH,		/* bar_type */
		2,			/* num_scales */
		update_metadata_bar,	/* update() */
		NULL,			/* init() */
      		"UFS Meta\nCache",	/* label */
					/* legend */
	 	{2, {"Hits", "Misses" } },
	},
	{
		LENGTH,		/* bar_type */
		5,			/* num_scales */
		update_pagefault_bar,	/* update() */
		NULL,			/* init() */
      		"Page\nFaults",		/* label */
					/* legend */
		{5, {"COW", "Page In", "Page Out", "React", "Zero" } },
	},
	{
		PROPORTION,		/* bar_type */
		1,			/* num_scales */
		update_swap_bar,	/* update() */
		NULL,			/* init() */
      		"Swap\nUsage",		/* label */
					/* legend */
		{2, {"Free", "Used"} },
	},
/*
** MKN: New entry for Intr bar
*/
	{
		LENGTH,
		3,
		update_intr_bar,
		NULL,
		"Intr\nActivity",
		{3,{"Device","Syscall","CtxSw"}},
	},
/*
** MKN: New entry for Syscall bar
*/
	{
		PROPORTION,
		1,
		update_syscall_bar,
		NULL,
		"Syscall\nActivity",
		{5,{"Read","Write","Exec","Fork+Vfork","Other"}},
	},
};

/*
** MKN: Definition of bar_info should go to "vmubc.h" to automate adjustment of MAX_BARS.
*/
#define NUM_BARS (sizeof(bar_info) / sizeof( BarInfo ))


static void init_cpu_bar( nbar )
{
    if (ncpus > 1)
	bar_info[nbar].bar_type = MULTI_PROPORTION;
    else
	bar_info[nbar].bar_type = PROPORTION;

    bar_info[nbar].num_scales = ncpus;
	
}

/* 
 *  These procedures gets called every interval to compute and display the
 *  bars. It should call draw_bar() with the bar number, the array of
 *  integer values to display in the bar, and the number of values in
 *  the array.
 */


static void update_cpu_bar(nbars)
int nbars;
{
	int states[5*(MAX_CPUS+1)];
	int i, j;
	
	if (table (TBL_SYSINFO, 0, (char *) &info, 1, sizeof (info)) < 0)
		perror ("sysinfo");
	
	else
		{
			int user   = info.si_user - prev_info.si_user;
			int system = info.si_sys  - prev_info.si_sys;
			int idle   = info.si_idle - prev_info.si_idle;
			int nice   = info.si_nice - prev_info.si_nice;
	      		int wait = info.wait - prev_info.wait;
			int total;

			total = user+system+idle+nice+wait;
			if (total == 0)
				total = 1;
			
			/* legend data */
			states[0] = (int)(0.5+100.0*(float)user/(float)total);
			states[1] = (int)(0.5+100.0*(float)nice/(float)total);
			states[2] = (int)(0.5+100.0*(float)system/(float)total);
/*
** MKN: Swap wait and idle
*/
			states[3] = (int)(0.5+100.0*(float)(wait)/(float)total);
			states[4] = (int)(0.5+100.0*(float)(idle)/(float)total);

			if (ncpus > 1) {
			    getallcpudata();
			    for (i = 0; i < ncpus; i++)
				for (j = 0; j < 5; j++)
				    states[ 5*(i+1) + j ] =
					cpu_info[i].ticks[j];
			}
			draw_bar (0, states, 5);
			
			prev_info = info;
		}
}
static void update_memory_bar(nbars)
int nbars;
{
	int states[5];
	
#if DUVER < 500
	if (vm_statistics(current_task(), &vmstat) != KERN_SUCCESS)
		perror ("vm_statistics");
	
#else  /* DUVER < 500 */
        if (table(TBL_VMSTATS, 0, &vmstat, 1, sizeof(vmstat)) != 1) {
                perror("tbl_vmstats could not be read");
        }
#endif /* DUVER < 500 */
	
	else
		{
/**
*** MKN: Correct V5 behaviour of memory bar, split active pages
*** MKN: On firstrun initialize prev_vmstat to avoid widget confusion 
**/
			if (firstrun) prev_vmstat = vmstat;
#if DUVER < 500
			vm_managed_pages = vmstat.active_count + vmstat.inactive_count
					 + vmstat.wire_count + vmstat.free_count;
			states[0] = vmstat.active_count - ubc_pages;
#else  /* DUVER < 500 */
			vm_managed_pages = vmstat.active_count + vmstat.inactive_count
					 + vmstat.wire_count + vmstat.ubc_pages + vmstat.free_count;
			ubc_pages = vmstat.ubc_pages;
                        states[0] = vmstat.active_count;
#endif /* DUVER < 500 */
			states[1] = vmstat.inactive_count;
			states[2] = vmstat.wire_count;
			states[3] = ubc_pages;
			states[4] = vmstat.free_count;
			draw_bar (1, states, 5);
	}
}

static void update_pagefault_bar(nbars)
int nbars;
{
	int states[5];
	
	/* Update page fault info */
	/* MKN: No need to read TBL_VMSTATS again */

	states[0] = vmstat.cow_faults - prev_vmstat.cow_faults;
	states[1] = vmstat.pageins - prev_vmstat.pageins;
	states[2] = vmstat.pageouts - prev_vmstat.pageouts;
	states[3] =
	    vmstat.reactivations - prev_vmstat.reactivations;
	states[4] =
	    vmstat.zero_fill_count-prev_vmstat.zero_fill_count;
	draw_bar( 5, states, 5 );

	prev_vmstat = vmstat;
}

static void update_UBC_usage_bar(nbars)
int nbars;
{
	int states[5];

#if DUVER >= 500
	/* MKN: No need to reread TBL_VMSTATS */
#else  /* DUVER >= 500 */
	if (lseek(kmem, (long) nl[N_UBC_PAGES].n_value, SEEK_SET) !=
	    (long) nl[N_UBC_PAGES].n_value) {
		perror("ubc_pages; lseek");
	} else
		if (read(kmem, &ubc_pages, sizeof(ubc_pages)) !=
		    sizeof(ubc_pages)) {
			perror("ubc_pages; read");
		}
	
	if (lseek(kmem, (long) nl[N_UBC_MINPAGES].n_value, SEEK_SET) !=
	    (long) nl[N_UBC_MINPAGES].n_value) {
		perror("ubc_minpages; lseek");
	} else
		if (read(kmem, &ubc_minpages, sizeof(ubc_minpages)) !=
		    sizeof(ubc_minpages)) {
			perror("ubc_minpages; read");
		}

	if (lseek(kmem, (long) nl[N_UBC_MAXPAGES].n_value, SEEK_SET) !=
	    (long) nl[N_UBC_MAXPAGES].n_value) {
		perror("ubc_maxpages; lseek");
	} else
		if (read(kmem, &ubc_maxpages, sizeof(ubc_maxpages)) !=
		    sizeof(ubc_maxpages)) {
			perror("ubc_maxpages; read");
		}
	if (lseek(kmem, (long) nl[N_UBC_DIRTYPAGES].n_value, SEEK_SET) !=
	    (long) nl[N_UBC_DIRTYPAGES].n_value) {
		perror("ubc_dirtypages; lseek");
	} else
		if (read(kmem, &ubc_dirtypages, sizeof(ubc_dirtypages)) !=
		    sizeof(ubc_dirtypages)) {
			perror("ubc_dirtypages; read");
		}
#endif /* DUVER >= 500 */

/**
*** MKN: Cleanup of calulation
***      Use ubc_minpercent and ubc_maxpercent to calculate minpages/maxpages
***/
#if DUVER >= 500
ubc_minpages = ((float)vm_managed_pages * ((float)ubc_minpercent/100.));
ubc_maxpages = ((float)vm_managed_pages * ((float)ubc_maxpercent/100.));
ubc_pages = vmstat.ubc_pages;
ubc_dirtypages = 0;
#endif /* DUVER >= 500 */
	states[0] = ubc_minpages;
	states[1] = ubc_pages-ubc_dirtypages;
	states[2] = ubc_dirtypages;
	states[3] = vm_managed_pages-ubc_pages;
	states[4] = vm_managed_pages-ubc_maxpages;

	draw_bar(2,states,5); 
}

static void update_UBC_cache_bar(nbars)
int nbars;
{
	int states[5];
	long hits=0,miss=0;

/**
** MKN: Use vmstat for hits/miss data. Now also counting AdvFS events.
**/
#if DUVER < 500
	if (lseek(kmem, (long) nl[N_UFS_GETAPAGE_STATS].n_value, SEEK_SET) !=
	    (long) nl[N_UFS_GETAPAGE_STATS].n_value) {
		perror("ufs_getapage_stats; lseek");
	} else
		if (read(kmem, &pg_stats, sizeof(pg_stats)) !=
		    sizeof(pg_stats)) {
			perror("ufs_getapage_stats; read");
		}

	    if (firstrun) old_pg_stats = pg_stats; // MKN: First-Pass initialization

	    hits = pg_stats.read_hits - old_pg_stats.read_hits;
	    miss = pg_stats.read_miss - old_pg_stats.read_miss;
	    old_pg_stats = pg_stats;
#else  /* DUVER < 500 */
	    hits = vmstat.hits - prev_vmstat.hits;
	    miss = (vmstat.lookups-vmstat.hits) - 
		   (prev_vmstat.lookups-prev_vmstat.hits);
#endif /* DUVER < 500 */
	    states[0] = hits;
	    states[1] = miss;

	    draw_bar(3,states,2); 

}

static void update_metadata_bar(nbars)
int nbars;
{
	int states[5];

	if (lseek(kmem, (long) nl[N_BIOSTATS].n_value, SEEK_SET) !=
	    (long) nl[N_BIOSTATS].n_value) {
		perror("biostats; lseek");
	} else
		if (read(kmem, &bstats, sizeof(bstats)) !=
		    sizeof(bstats)) {
			perror("biostats; read");
		}

	    if (firstrun) old_bstats = bstats; // MKN: First-Pass initialization

	    states[0] = bstats.getblk_hits - old_bstats.getblk_hits;
	    states[1] = bstats.getblk_misses - old_bstats.getblk_misses;

	    old_bstats = bstats;
	    draw_bar(4,states,2); 
}

static void update_swap_bar(nbars)
int nbars;
{
	int states[5];
	struct tbl_swapinfo swapinfo;


	if (table( TBL_SWAPINFO, -1, &swapinfo, 1, sizeof(struct tbl_swapinfo)) < 0) {
	    perror("table: TBL_SWAPINFO");
	}
	states[0] = swapinfo.free;
	states[1] = swapinfo.size - swapinfo.free;

#ifdef OLDSWAP
	/* Swap usage */
	if (lseek(kmem, (long) nl[N_VM_TOTAL_SWAP_SPACE].n_value, SEEK_SET) !=
	    (long) nl[N_VM_TOTAL_SWAP_SPACE].n_value) {
		perror("vm_total_swap_space; lseek");
	} else
		if (read(kmem, &vm_total_swap_space, sizeof(vm_total_swap_space)) !=
		    sizeof(vm_total_swap_space)) {
			perror("vm_total_swap_space; read");
		}
	if (lseek(kmem, (long) nl[N_VM_SWAP_SPACE].n_value, SEEK_SET) !=
	    (long) nl[N_VM_SWAP_SPACE].n_value) {
		perror("vm_swap_space; lseek");
	} else
		if (read(kmem, &vm_swap_space, sizeof(vm_swap_space)) !=
		    sizeof(vm_swap_space)) {
			perror("vm_swap_space; read");
		}

	states[0] = vm_swap_space;
	states[1] = vm_total_swap_space - vm_swap_space;
#endif
	draw_bar(6, states, 2);

}

/*
** MKN: New routine for INTR. Values scaled to intervall (1/sec)
*/
static void update_intr_bar(nbars)
int nbars;
{
  int states[3];

	if (table( TBL_INTR, 0, &intrtb, 1, sizeof(struct tbl_intr)) < 0) {
		perror(" table: TBL_INTR");
	}
	else {
		if (firstrun) prev_intrtb = intrtb;
		states[0] = (int)(0.5+(float)(intrtb.in_devintr - prev_intrtb.in_devintr)/(float)interval);
		states[1] = (int)(0.5+(float)(intrtb.in_syscalls - prev_intrtb.in_syscalls)/(float)interval);
		states[2] = (int)(0.5+(float)(intrtb.in_context - prev_intrtb.in_context)/(float)interval);
		draw_bar(7, states, 3);
	}
}

/*
** MKN: New routine for Syscalls. Values scaled to intervall (1/sec)
*/
static void update_syscall_bar(nbars)
int nbars;
{
  int states[5];

	if (table( TBL_SCALLS, 0, &scalls, 1, sizeof(struct tbl_scalls)) < 0) {
		perror(" table: TBL_SCALLS");
	}
	else {
		if (firstrun) prev_scalls = scalls;
		states[0] = (int)(0.5+(float)(scalls.sysread - prev_scalls.sysread)/(float)interval);
		states[1] = (int)(0.5+(float)(scalls.syswrite - prev_scalls.syswrite)/(float)interval);
		states[2] = (int)(0.5+(float)(scalls.sysexec - prev_scalls.sysexec)/(float)interval);
		states[3] = scalls.sysfork - prev_scalls.sysfork 
			  + intrtb.in_vforks - prev_intrtb.in_vforks;
		states[3] = (int)(0.5+(float)states[3]/(float)interval);
		states[4] = scalls.syscall - scalls.sysread - scalls.sysfork
			  - scalls.syswrite - scalls.sysexec - (prev_scalls.syscall
			  - prev_scalls.sysread - prev_scalls.sysfork 
			  - prev_scalls.syswrite - prev_scalls.sysexec);
		states[4] = (int)(0.5+(float)states[4]/(float)interval);
		draw_bar(8, states, 5);
	}
	prev_scalls = scalls;
	prev_intrtb = intrtb;
	firstrun = 0; /* MKN: This line should be moved to last bar routine */
}



struct machine_slot *mslot;
#define NCPUS MAX_CPUS
#define KERNEL "/vmunix"
#define KCOREF "/dev/kmem"
#define NSLOT  MAX_CPUS
struct machine_slot machine_slot[NSLOT];

/* read kmem at addr into to, nbytes bytes */
static int getkmem(addr,to,nbytes,varname)
long addr, nbytes;
char *to, *varname;
{
  if (lseek(kmem,addr,0) != -1) {
    if (nbytes == read(kmem,to,nbytes)) return(1);
  }
  if (varname != 0L) {
    fprintf(stderr,"\nkmem - cannot read %s\n",varname);
    exit(1);
  } else {
    return(0);
  }
}

static char kernelName[1024]="/";
static char *get_kernel_name()
{
    if (strlen(kernelName) <= 1) {
        getsysinfo( GSI_BOOTEDFILE, kernelName+1, 1024, 0, 0);
    }
    return(kernelName);
}

static int initkmem()
{
  struct nlist nlis[4];
  int *ncp;

  struct cpu_info CpuInfo;
  
  getsysinfo( GSI_CPU_INFO, &CpuInfo, sizeof( struct cpu_info ), 0, NULL );
#if DUVER < 500
  nlis[0].n_name = "_machine_slot";
  nlis[1].n_name = NULL;
  if(nlist(get_kernel_name(),nlis) < 0 ) {
    fprintf(stderr, "nlist failed to read _machine_slot:\n");
    exit(1);
  }
  mslot = (struct machine_slot *)nlis[0].n_value;
#endif /* DUVER < 500 */

#ifdef GSI_CPUS_IN_BOX
  getsysinfo( GSI_CPUS_IN_BOX, &ncpus, sizeof( ncpus ), 0, NULL  );
#else
  ncpus = 1;
#endif

#ifdef OLDCPUS
  ncp = (int *)nlis[1].n_value;
  getkmem(ncp,&ncpus,sizeof(ncpus),"ncpus");
#endif

  if (ncpus > NCPUS)
    ncpus = NCPUS;

}

static long cps_time[NCPUS][CPU_STATE_MAX];
static long old_cps_time[NCPUS][CPU_STATE_MAX];

#ifndef CPU_STATE_USER
# define CPU_STATE_USER 0
#endif
#ifndef CPU_STATE_NICE
# define CPU_STATE_NICE 1
#endif
#ifndef CPU_STATE_SYSTEM
# define CPU_STATE_SYSTEM 2
#endif
#ifndef CPU_STATE_IDLE
# define CPU_STATE_IDLE 3
#endif
#ifndef CPU_STATE_WAIT
# define CPU_STATE_WAIT 4
#endif

static int getcpudata( int cpu_num, struct tick_struct *cpu_info )
{
  register int i; /* machine slot */
  register int j; /* CPU number */

  for (i=0,j=-1;i<NSLOT; i++) {
    if ( machine_slot[i].is_cpu != 1 ) continue;

    j++;  /* It's a CPU */


    if ( j == cpu_num ) {
        if ( !machine_slot[i].running ) {
	  return(0);
        }

          cpu_info->ticks[0] = machine_slot[i].cpu_ticks[CPU_STATE_USER];
          cpu_info->ticks[1] = machine_slot[i].cpu_ticks[CPU_STATE_NICE];
          cpu_info->ticks[2] = machine_slot[i].cpu_ticks[CPU_STATE_SYSTEM];
/*
** MKN: Swap wait and idle
*/
          cpu_info->ticks[3] = machine_slot[i].cpu_ticks[CPU_STATE_WAIT];
          cpu_info->ticks[4] = machine_slot[i].cpu_ticks[CPU_STATE_IDLE];
	  return(1);
    }
  }
}

static int getallcpudata()
{
  int i, k;
  float total;
  struct tick_struct temp_info;

#if DUVER < 500
  getkmem(mslot,machine_slot,sizeof(machine_slot),"machine_slot");
#else  /* DUVER < 500 */
        if (table(TBL_MACHINE_SLOT, 0, machine_slot, 1,
                sizeof(struct machine_slot) * ncpus) == -1) {
                perror("table(TBL_MACHINE_SLOT) error");
                exit(1);
        }
#endif /* DUVER < 500 */


  for (i = 0; i < ncpus; i++) {
    if( getcpudata( i, &cpu_info[i] ) ) {
      temp_info.ticks[0] = cpu_info[i].ticks[0] - prev_cpu_info[i].ticks[0];
      temp_info.ticks[1] = cpu_info[i].ticks[1] - prev_cpu_info[i].ticks[1];
      temp_info.ticks[2] = cpu_info[i].ticks[2] - prev_cpu_info[i].ticks[2];
      temp_info.ticks[3] = cpu_info[i].ticks[3] - prev_cpu_info[i].ticks[3];
      temp_info.ticks[4] = cpu_info[i].ticks[4] - prev_cpu_info[i].ticks[4];
    }
    else {
      temp_info.ticks[0] = 0;
      temp_info.ticks[1] = 0;
      temp_info.ticks[2] = 0;
      temp_info.ticks[3] = 0;
      temp_info.ticks[4] = 0;
    }
    total = temp_info.ticks[0] +
	    temp_info.ticks[1] +
	    temp_info.ticks[2] +
	    temp_info.ticks[3] +
	    temp_info.ticks[4];

    if (total <= 0) {
	cpu_info[i].ticks[0] = 0;
	cpu_info[i].ticks[1] = 0;
	cpu_info[i].ticks[2] = 0;
	cpu_info[i].ticks[3] = 0;
	cpu_info[i].ticks[4] = 0;
        prev_cpu_info[i] = cpu_info[i];
    }
    else {
        for (k = 0; k < 5; k++)
	    temp_info.ticks[k] = (0.5 + 100.0 * (float)temp_info.ticks[k])/total;
        prev_cpu_info[i] = cpu_info[i];
        cpu_info[i] = temp_info;
    }
  }
}

/* 
** This exists only to catch errors
*/
static void bad_func()
{
    fprintf(stderr,"Bad update function\n");
}

/*
**
** Externally visible entry points
*/

/* Called at the beginning to inquire how many bars are needed. */
int
num_bars()
{
  return NUM_BARS;
}


/* Called after num_bars to ask for the bar names */
char * label_bar(nbar)
     int nbar;
{
  if (nbar >= NUM_BARS || nbar < 0)
	return NULL;
  else
      return bar_info[nbar].label;
}


int bar_type( nbar )
    int nbar;
{
    if (nbar >= NUM_BARS) {
	    fprintf( stderr, "Bad bar number in routine bar_type()\n");
	    return (-1);
    }
    if (nbar == 0) {
	    if (ncpus > 1)
		return( MULTI_PROPORTION );
	    else
		return( PROPORTION );
    }
    return( bar_info[nbar].bar_type );
}

int num_scales( nbar )
int nbar;
{
    return( bar_info[nbar].num_scales );
}
int num_legends( nbar )
    int nbar;
{
    return( bar_info[nbar].legend.num_labels );
}

char * legend_name( nbar, nkey )
	int nbar;
	int nkey;
{
    return( bar_info[nbar].legend.labels[nkey] );
}

/* 
 *  Called before the bar widgets are created to perform any machine dependent
 *  initializations.
 */
void init_bars()
{
    struct utsname name;
    int i;
    char *kname = get_kernel_name();

	if (nlist(kname, nl) <0) {
		perror(kname);
		exit(1);
	}

	if ((kmem = open(_PATH_KMEM, O_RDONLY)) < 0) {
		perror(_PATH_KMEM);
		exit(1);
	}
#if DUVER < 500
	else if (lseek(kmem, (long) nl[N_UFS_GETAPAGE_STATS].n_value, SEEK_SET) ==
	    (long) nl[N_UFS_GETAPAGE_STATS].n_value) {
		(void)read(kmem, &old_pg_stats, sizeof(old_pg_stats));
	}
#endif /* DUVER < 500 */	
	else if (lseek(kmem, (long) nl[N_BIOSTATS].n_value, SEEK_SET) ==
	    (long) nl[N_BIOSTATS].n_value) {
		(void)read(kmem, &old_bstats, sizeof(old_bstats));
	}

/*
** MKN: Initialize firstrun. Used to avoid huge numbers being fed into
**      widgets first time after start/resume/intervall-change
*/
	firstrun = 1;

	bzero( &vmstat, sizeof( vmstat ) );
	prev_vmstat = vmstat;

/*
** MKN: Initialize INTR structure
*/
	bzero( &intrtb, sizeof( intrtb ) );
	prev_intrtb = intrtb;
/*
** MKN: Initialize Syscall structure
*/
	bzero( &scalls, sizeof( intrtb ) );
	prev_scalls = scalls;

        if (uname( &name ) < 0) {
	    perror("uname(2) failed!");
	    exit(1);
        }
	if (!strncmp(name.release, "V2.", 3)) {
	    /* V2.x */
	    ncpus = 1;
	}
	else {
	    /* assume V3.0 or later */
            initkmem();

	    if (ncpus > 1) {
		cpu_info = (struct tick_struct *)malloc( ncpus * sizeof( struct tick_struct ));
		prev_cpu_info = (struct tick_struct *)malloc( ncpus * sizeof( struct tick_struct ));

		/* Prime that data structures */

		bzero( (char *)prev_cpu_info,
			ncpus * sizeof( struct tick_struct ) );
	        getallcpudata();
	    }
	    else {
		cpu_info = NULL;
		prev_cpu_info = NULL;
	    }
	}

    for (i = 0; i < num_bars(); i++) {
	if (bar_info[i].init != NULL)
	    (*bar_info[i].init)(i);
    }
	
/**
*** MKN: Query VM subset for ubc_minpercent, ubc_maxpercent
**/
#if DUVER >= 500
	strcpy(ubc_attr[0].name,"ubc-minpercent");
	strcpy(ubc_attr[1].name,"ubc_minpercent");
	strcpy(ubc_attr[2].name,"ubc-maxpercent");
	strcpy(ubc_attr[3].name,"ubc_maxpercent");

	cfg_retval = cfg_subsys_query(NULL, "vm", ubc_attr, 4);
	if ( cfg_retval != CFG_SUCCESS)
	     perror("cfg_subsys_query failed");
	else {
/**
*** MKN: Check for both V4 an V5 names ('-' vs. '_'). V5 value will winn if defined.
**/
		if(ubc_attr[0].status == CFG_ATTR_SUCCESS) ubc_minpercent = ubc_attr[0].attr.num.val; 
		if(ubc_attr[1].status == CFG_ATTR_SUCCESS) ubc_minpercent = ubc_attr[1].attr.num.val; 
		if(ubc_attr[2].status == CFG_ATTR_SUCCESS) ubc_maxpercent = ubc_attr[2].attr.num.val; 
		if(ubc_attr[3].status == CFG_ATTR_SUCCESS) ubc_maxpercent = ubc_attr[3].attr.num.val; 
	}
#endif 

}

void *update_func(nbar)
int nbar;
{
    switch (nbar) {
	case 0 : 
		return( update_cpu_bar); break;
	case 1 : 
		return( update_memory_bar); break;
	case 2 : 
		return( update_UBC_usage_bar); break;
	case 3 : 
		return( update_UBC_cache_bar); break;
	case 4 : 
		return( update_metadata_bar); break;
	case 5 : 
		return( update_pagefault_bar); break;
	case 6 :
		return( update_swap_bar); break;
	case 7 : /* MKN: Add callback for intr bar */
		return( update_intr_bar); break;
	case 8 : /* MKN: Add callback for syscall bar */
		return( update_syscall_bar); break;
	default :
	    return( bad_func );
	    break;
    }
}
