/*
             Copyright 1994,1995 Digital Equipment Corporation
	     Copyright 2000 Compaq Computer Corporation
                        All rights reserved.

   Redistribution and use in source and binary forms are
   permitted provided that: 1) source distributions retain this
   entire copyright notice and 2) distributions including
   binaries display the following acknowledgment in its
   documentation:

           "This product includes software developed by the
            Digital Equipment Corporationi and Compaq Computer
	    Corporation" 
*/
/*
** Author:  George Chaltas
**	    Digital Equipment Corporation
** fixed memory display for very large memory (128GB) by Bernard Laforgue, Compaq
** updated for Tru64 UNIX V5.x by Bernard Laforgue, Compaq
** updated for Tru64 UNIX V5.x by Martin Knoblauch (MKN), Compaq
*/
#include <stdio.h>
#include <Mrm/MrmAppl.h>
#include <Xm/MessageB.h>
#include "vmubc.h"

#ifndef UIDFILE
#define UIDFILE "vmubc.uid"
#endif

static char *version_string = "VmUBC V5.0 (13-Dec-2000)";
static int nbars;

static char ** formats, ** kformats;
static int *format_threshold;
static char header1[350]; // MKN: Increased to 192 for "-intr"
static char header2[350]; // MKN: Increased to 192 for "-intr"
static XtAppContext app_context;

Pixel black, red, green, blue, cyan, magenta, yellow, white;
int *bar_height;
int *bar_width;
GC gc;


#define MAX_ROWS MAX_CPUS

typedef struct {
	Widget box_widget;
	Widget bar_widget;
	int num_legends;
	Widget legend[MAX_VALUES];
	int num_values;
	int bar_type;  /* PROPORTION, MULTI_PROPORTION or  LENGTH */
	int values[MAX_VALUES];
	int num_scales;
	int multi_values[MAX_ROWS][MAX_VALUES];
	void (*update)();
} Bar; 

#ifndef MAX
#define MAX( A, B ) (((A) < (B)) ? (B) : (A))
#endif


Bar bars[MAX_BARS];

Widget	topW = NULL,
	scale_boxW = NULL,
	scaleW = NULL,
    	mainW = NULL,
	copyrightW = NULL,
	suspendW = NULL;
	

MrmHierarchy Hier;
MrmCode	    dumcls;

static char *dbfiles[] = {UIDFILE};


XtCallbackProc control_cb();
void expose_cb();
void create_cb();
void create_label_cb();
void create_bar_cb();

Arg args[10];

MRMRegisterArg		 reglist[] = 
	{ 
		{"control_cb",		(caddr_t)control_cb},
		{"expose_cb",		(caddr_t)expose_cb},
		{"create_cb",		(caddr_t)create_cb},
		{"create_label_cb",	(caddr_t)create_label_cb},
		{"create_bar_cb",	(caddr_t)create_bar_cb},
	} ;

#define NUM_REGNAMES ( sizeof( reglist )/ sizeof( MRMRegisterArg ) )

extern int log_to_tty;
extern int do_bar[];
extern int initialize;

extern struct def_t defs;
extern int interval;
extern int firstrun; // MKN: First pass initialization flag
extern int count;

/* Application Resources - no particular widget */
static XtResource application_resources[] = {
    {"interval", "Interval", XtRInt, sizeof(int),
	    XtOffset(struct def_t *,interval),
	    XtRImmediate,(XtPointer)DEF_INTERVAL},
    {"count", "Count", XtRInt, sizeof(int),
	    XtOffset(struct def_t *,count),
	    XtRImmediate,(XtPointer)DEF_COUNT},
    {"hidebars", "HideBars", XtRInt, sizeof(int),
	    XtOffset(struct def_t *, hidebars),
	    XtRBool,FALSE},
};

/*
*  Command line options table. The command line is parsed for these,
*  and it sets/overrides the appropriate values in the resource
*  database
*/

static XrmOptionDescRec optionDescList[] = {
    {"-interval",   ".interval",    XrmoptionSepArg,    (caddr_t) NULL},
    {"-count",	    ".count",	    XrmoptionSepArg,    (caddr_t) NULL},
    {"-hidebars",   ".hidebars",    XrmoptionNoArg,	(caddr_t) "1"},
};

XtIntervalId timer = NULL;

static int bars_done; //MKN.
init_for_tty()
{
    int i;

    /* Turn off buffering of stdout */

    setbuf( stdout, NULL );


    /* Determine formats */

    nbars = num_bars();
    bars_done = 0;

    formats = (char **)malloc( nbars * sizeof( char * ));
    kformats = (char **)malloc( nbars * sizeof( char * ));
    format_threshold = (int *)malloc( nbars * sizeof( int ));
    for (i = 0; i < nbars; i++) {
	bars[i].update = (void *)update_func( i );
	bars[i].num_legends = num_legends( i );
	bars[i].bar_type = bar_type( i );
	if (!do_bar[i]) continue;
	/* Format strings */
	switch(i) {
	    case 0 : formats[i] = "%3d"; kformats[i] = "%2d%c";
		     /* format_threshold not used */
		     format_threshold[i] = 1000; break;
	    case 1 :
	    case 2 : 
	    case 3 :
	    case 4 :
	    case 6 :
	    case 7 : /* MKN: INTR bar. Use same format as paging */
	    	     formats[i] = "%5d"; kformats[i] = "%4d%c";
		     format_threshold[i] = 10000; break;
	    case 5 :
	    case 8 : /* MKN: Syscall bar. Use same format as paging */
	    	     formats[i] = "%4d"; kformats[i] = "%3d%c";
		     format_threshold[i] = 1000; break;
	    default :
		fprintf(stderr, "Unrecognized bar number %d\n", i );
		exit(-1);
	}
        /* header String */
	/* MKN: Some adjustments for "nicer" alignment */
	switch(i) {
	    case 0 :
		/* 16 */
//------------------------------1234567890123456789012345678901234567890
		strcat(header1,"|     CPU %     ");
		strcat(header2,"|  U  N  S  W  I");
		break;
	    case 1 :
		/* 31 */
		strcat(header1,"|       Memory Pages      ");
		strcat(header2,"|  Act Inac Wire  UBC Free");
		break;
	    case 2 :
		/* 25 */
		strcat(header1,"|    UBC Page Allocation  ");
		strcat(header2,"|  Min  Cln Drty   VM-Only");
		break;
	    case 3 :
		/* 12 */
	        strcat(header1,"| UBC Cache");
		strcat(header2,"|  Hit Miss");
		break;
	    case 4 :
		/* 11 */
	        strcat(header1,"| Metadata ");
		strcat(header2,"|  Hit Miss");
		break;
	    case 5 :
		/* 26 */
		strcat(header1,"|     Page Faults    ");
		strcat(header2,"| COW pin pou rct zer");
		break;
	    case 6 :
		/* 13 */
		strcat(header1,"|    Swap  ");
		strcat(header2,"| Free Used");
		break;
	    case 7 : /* MKN: Intr bar */
	    	/* 16 */
		strcat(header1,"| INTR Activity ");
		strcat(header2,"|  Dev  Sys CtSw");
		break;
	    case 8 : /* MKN: Syscall bar */
	    	/* 21 */
		strcat(header1,"| Syscall Activity   ");
		strcat(header2,"|  Rd  Wr  Ex  Fo  Ot");
		break;
	    default :
		fprintf(stderr, "Unrecognized bar number (header) %d\n", i );
		exit(-1);
	}
    }
    strcat(header1,"|");
    strcat(header2,"|");
    }

write_tty_header()
{
    fprintf(stdout,"\n%s\n%s\n", header1, header2 );
}


create_widgets( progname, argc, argv )
char *progname;
int argc;
char	**argv;
{
	int	 i,db;
	XrmValue rv;
	Pixmap icon, create_icons();
	Display *dpy;
	MrmCode literal_type;
	Arg args[2];

        nbars = num_bars();;

	MrmInitialize();

	XtToolkitInitialize();
	app_context = XtCreateApplicationContext();

	dpy = XtOpenDisplay( app_context, NULL, progname, "VmUBC",
			    optionDescList, XtNumber(optionDescList),
			    &argc, argv);


	if (dpy == NULL) {
	    fprintf(stderr,"VmUBC: unable to open DISPLAY, using tty mode\n");
	    XtDestroyApplicationContext( app_context );
	    log_to_tty = TRUE;
	    init_for_tty();
	    return;
	}

	topW = XtAppCreateShell( progname, "VmUBC", 
			applicationShellWidgetClass,
			dpy, NULL, 0 );

/*
    if (argc > 1)
	usage();
*/
	

        XtGetApplicationResources(topW, &defs, application_resources,
			      XtNumber(application_resources), NULL, 0 );


/*
        icon = create_icons();
	XtSetArg( args[0], XtNiconPixmap, icon );

*/

	XtSetArg( args[0], XtNallowShellResize, TRUE );
	XtSetValues( topW, args, 1 );

	if(MrmOpenHierarchy  (1,&dbfiles[0],NULL,&Hier)!=MrmSUCCESS) {
		exit(1);
	}

	if(MrmRegisterNames(reglist,NUM_REGNAMES)!=MrmSUCCESS) {
		fprintf(stderr,"can't register procedures\n");
		exit(1);
	}

	if (MrmFetchColorLiteral(Hier,"k_black", XtDisplay(topW), NULL, &black)
	    != MrmSUCCESS) {
		fprintf(stderr,"can't find black in vmubc.uid\n");
	}

	if (MrmFetchColorLiteral(Hier,"k_red", XtDisplay(topW), NULL, &red)
	    != MrmSUCCESS) {
		fprintf(stderr,"can't find red in vmubc.uid\n");
	}

	if (MrmFetchColorLiteral(Hier,"k_green", XtDisplay(topW), NULL, &green)
	    != MrmSUCCESS) {
		fprintf(stderr,"can't find green in vmubc.uid\n");
	}

	if (MrmFetchColorLiteral(Hier,"k_yellow", XtDisplay(topW), NULL,&yellow)
	    != MrmSUCCESS) {
		fprintf(stderr,"can't find yellow in vmubc.uid\n");
	}

	if (MrmFetchColorLiteral(Hier,"k_blue", XtDisplay(topW), NULL, &blue)
	    != MrmSUCCESS) {
		fprintf(stderr,"can't find blue in vmubc.uid\n");
	}

	if (MrmFetchColorLiteral(Hier,"k_cyan", XtDisplay(topW), NULL, &cyan)
	    != MrmSUCCESS) {
		fprintf(stderr,"can't find cyan in vmubc.uid\n");
	}

	if (MrmFetchColorLiteral(Hier,"k_magenta", XtDisplay(topW), NULL, &magenta)
	    != MrmSUCCESS) {
		fprintf(stderr,"can't find magenta in vmubc.uid\n");
	}

	if (MrmFetchColorLiteral(Hier,"k_white", XtDisplay(topW), NULL, &white)
	    != MrmSUCCESS) {
		fprintf(stderr,"can't find white in vmubc.uid\n");
	}

        if (MrmFetchLiteral(Hier,"k_height", XtDisplay(topW),
	     &bar_height, &literal_type) != MrmSUCCESS) {
		exit(1);
	}
        if (MrmFetchLiteral(Hier,"k_width", XtDisplay(topW),
	     &bar_width, &literal_type) != MrmSUCCESS) {
		exit(1);
	}

	if(MrmFetchWidget(Hier,"main",topW, &mainW,&dumcls) != MrmSUCCESS) 
	{
		fprintf(stderr,"Can not fetch main Widget \n");
		exit(1);
	}

	XtManageChild(mainW);
	XtRealizeWidget(topW);

        gc = XCreateGC( XtDisplay(topW), XtWindow(topW), 0, NULL );

}


/*
** Callbacks
*/
/*
** Creation Callbacks
*/
static int create_bar = -1;
static int create_legend = -1;


void build_tool_menu( w )
Widget		w;
{
    int i;
    Widget tool_button;
    Arg arg[5];
    int arg_num;
    XmString button_label;
    int *callback_values;

    /* DO NOT FREE THIS!! */
    callback_values = (int *)malloc(nbars*sizeof(int));

    for (i = 0; i < nbars; i++) {
	button_label = XmStringCreate( label_bar( i ), XmSTRING_DEFAULT_CHARSET );
	callback_values[i] = i;
	arg_num = 0;
	XtSetArg( arg[arg_num], XmNlabelString, button_label ); arg_num++;
	if (!do_bar[i]) {
	    XtSetArg( arg[arg_num], XmNset, FALSE ); arg_num++;
	}
	if (MrmFetchWidget(Hier,"tool_button", w,
	    &tool_button, &dumcls) != MrmSUCCESS)
	{
	    fprintf(stderr,"Can not fetch tool_button Widget %d\n", i);
	    exit(1);
	}

	XtSetValues( tool_button, arg, arg_num );
	XtAddCallback( tool_button, XmNvalueChangedCallback, control_cb, &callback_values[i] );
	XtManageChild( tool_button );
	XtFree( button_label );
    }
}

void create_cb( w, tag, reason )
Widget		w;
int		*tag;
unsigned long	*reason;
{
    int i;
    Arg arg[10];
    Widget bar_box, legend_box;

#ifdef DEBUG
	fprintf(stderr,"Bar Box %d created\n", *tag);
#endif
	
    switch ( *tag ) {
    case 0 :
        /* fetch bar boxes */
        for (i = 0; i < nbars; i++) {
	    create_bar = i;
	    if(MrmFetchWidget(Hier,"bar_box",w, 
		&bar_box, &dumcls) != MrmSUCCESS) 
	    {
		fprintf(stderr,"Can not fetch bar_box Widget %d\n", i);
		exit(1);
	    }
	    if (do_bar[i])
	        XtManageChild(bar_box);
	    bars[i].box_widget = bar_box;
        }
        break;
    case 1 :
	bars[create_bar].update = (void *)update_func( create_bar );
	bars[create_bar].num_legends = num_legends( create_bar );
	bars[create_bar].bar_type = bar_type( create_bar );
	    for (i = 0; i < bars[create_bar].num_legends; i++) {
	        create_legend = i;
	        if(MrmFetchWidget(Hier,"legend_box",w, 
		    &legend_box, &dumcls) != MrmSUCCESS) 
	        {
		    fprintf(stderr,"Can not fetch legend_box Widget %d\n", i);
	    	    exit(1);
	        }
	        XtManageChild(legend_box);
	    }
	break;
    case k_Interval :
	scaleW = w;
	break;
    case k_Freeze :
	suspendW = w;
	break;
    case k_tool_menu :
	build_tool_menu( w );
    }

}
void set_suspend_button( state )
Boolean state;
{
    XmToggleButtonSetState( suspendW, state, TRUE );
}
void create_label_cb( w, tag, reason )
Widget		w;
int		*tag;
unsigned long	*reason;
{
    switch (*tag) {

	case -1 :
	    /* set label string */
            set_label( w, label_bar(create_bar) );
	    break;
	case -2 :
            set_label( w, legend_name(create_bar, create_legend) );
	    switch (create_legend) {
	        case 0 : set_colors( w, green, black ); break;
	        case 1 : set_colors( w, yellow, black ); break;
	        case 2 : set_colors( w, red, black ); break;
/*
** MKN: switch magenta and blue. Adaption for memory bar change
*/
	        case 3 : set_colors( w, magenta, black ); break;
	        case 4 : set_colors( w, blue, black ); break;
	    }
	    break;
	case -3 :
	    bars[create_bar].legend[create_legend] = w;
	    break;
	default :
	    fprintf(stderr,"Illegal tag %d to create_label_cb\n", *tag);
	    break;
    } 

        return;
}

void create_bar_cb( w, tag, reason )
Widget		w;
int		*tag;
unsigned long	*reason;
{
    Arg args[2];
    int i;
    unsigned int h;

#ifdef DEBUG
	fprintf(stderr, "Bar %d created\n", create_bar );
#endif
    bars[create_bar].bar_widget = w;
    bars[create_bar].num_values = 0;
    bars[create_bar].num_scales = num_scales( create_bar );

    /* adjust height of bar */
    h = bars[create_bar].num_scales * *bar_height;
    /* should subtract 1 for multi-cpus! */

    i = 0;
    XtSetArg(args[i], XmNheight, h); i++;
    XtSetValues( w, args, i );
}

/*
** Exposure Callbacks
*/
void expose_cb( w, tag, reason )
Widget		w;
int		*tag;
unsigned long	*reason;
{
    int i;
    void redraw_bar();
 
    for (i = 0; i < nbars; i++)
	if (bars[i].bar_widget == w) {
#ifdef DEBUG
	fprintf(stderr, "Bar %d exposed\n", i);
#endif
	redraw_bar( i );
	return;
	}
}


void clear_timer()
{
    if (timer == NULL)
	return;

    XtRemoveTimeOut( timer );
    timer = NULL;
}

XtCallbackProc control_cb( w, tag, reason )
Widget		w;
int		*tag;
unsigned long	*reason;
{
    int i;
    char buff[80];
    Widget t;
    Arg args[8];


    switch (*tag) {
        case k_exit :
		exit(0);
		break;
	case k_Freeze :
		if (timer == NULL) {
		    start_timer();
                    set_label( w, "Suspend Updates" );
		    firstrun = 1; // MKN: Do another first-pass after resuming update
		}
		else {
		    clear_timer();
                    set_label( w, "Resume Updates" );
		}
		break;
	case k_OK :	/* read scale */
		(void)XmScaleGetValue( scaleW, &interval );
		if (timer != NULL) {
		    clear_timer();
		    start_timer();
		}
		firstrun = 1; // MKN: Do a first-pass after changeing intervall
		break;
	case k_Interval :
		if (scale_boxW == NULL) {
		    if(MrmFetchWidget(
		        Hier,"Interval_dialog",topW, &scale_boxW,&dumcls)
		        != MrmSUCCESS) 
			    exit(1);
		}
		(void)XmScaleSetValue( scaleW, interval );
		XtManageChild( scale_boxW );
		break;
	case k_copyright :
		if (copyrightW == NULL) {
    		    XmString      vstring;
	
		    if(MrmFetchWidget(
		        Hier,"copyright",topW, &copyrightW,&dumcls)
		        != MrmSUCCESS) 
			    exit(1);
    			vstring = XmStringCreate(version_string, XmSTRING_DEFAULT_CHARSET);
			XtSetArg( args[0], XmNdialogTitle, vstring );
			XtSetValues( copyrightW, args, 1 );
    			XmStringFree(vstring);
			t = XmMessageBoxGetChild( copyrightW, XmDIALOG_CANCEL_BUTTON );
			XtUnmanageChild( t );
			t = XmMessageBoxGetChild( copyrightW, XmDIALOG_HELP_BUTTON );
			XtUnmanageChild( t );
		}
		XtManageChild( copyrightW );
		break;
	default :
	    if (XtIsManaged( bars[*tag].box_widget)) {
		XtUnmanageChild( bars[*tag].box_widget );
	    }
	    else {
		XtManageChild( bars[*tag].box_widget );
	    }
	    break;
    }
}

set_colors( w, fg, bg )
Widget w;
Pixel fg;
Pixel bg;
{
    int		  i;
    Arg           args[2];

    i = 0;
    XtSetArg(args[i], XmNforeground, fg); i++;
    XtSetArg(args[i], XmNbackground, bg); i++;
    XtSetValues( w, args, i );
}

/* This will recognise newline characters and act appropriately */
set_label( w, string )
Widget w;
char *string;
{
    int		  i;
    Arg           args[2];
    XmString      tcs;

    tcs = XmStringLtoRCreate(string, XmSTRING_DEFAULT_CHARSET);
    i = 0;
    XtSetArg(args[i], XmNlabelString, tcs); i++;
    XtSetValues( w, args, i );
    XmStringFree(tcs);
}

extern int bars_counted;
void write_bar( bar_num, states, nstates)
int bar_num;
int states[];
int nstates;
{
    register int i;

    fprintf(stdout,"|"); // MKN.
/*
** MKN: User uppercase letters M/K for binary units
*/
    for (i = 0; i < nstates; i++)
	if (states[i] >= (1024*format_threshold[bar_num]))
	    /* Use m notation */
    	    fprintf(stdout,kformats[bar_num],(states[i] >> 20), 'M');
	else if (states[i] >= format_threshold[bar_num])
	    /* Use k notation */
    	    fprintf(stdout,kformats[bar_num],(states[i] >> 10), 'K');
	else
    	    fprintf(stdout,formats[bar_num],states[i]);
    	    
   
/*
** MKN: Print final "|" in tty mode
*/
    bars_done++;
    if (bars_done == bars_counted) {fprintf(stdout,"|");bars_done=0;}
}

void
draw_bar(bar_num, states, nstates)
int bar_num;
int states[];
int nstates;
{
    int i, j;
    char buff[128];
    unsigned change_bar = FALSE;
    void redraw_bar();

    if (log_to_tty) {
	if (do_bar[bar_num] && !initialize)
	    write_bar(bar_num, states, nstates);
	return;
    }

    if (nstates != bars[bar_num].num_values) {
	change_bar = TRUE;
        bars[bar_num].num_values = nstates;
    }

    /* copy legend data */

    for (i = 0; i < nstates; i++) {
	if (states[i] != bars[bar_num].values[i]) {
	    bars[bar_num].values[i] = states[i];
	    change_bar = TRUE;
	}
    }

    if (bars[bar_num].bar_type == MULTI_PROPORTION) {
	/* copy bar data */

	for (i = 0; i < bars[bar_num].num_scales; i++)
	    for (j = 0; j < bars[bar_num].num_values; j++)
		bars[bar_num].multi_values[i][j] =
			states[ (i+1)*nstates + j ];
	    
	change_bar = TRUE;
    }
	
    if (change_bar) {
	/* update the legends */
	for (i = 0; i < bars[bar_num].num_legends; i++) {
	    sprintf(buff, "%d", states[i] );
	    set_label( bars[bar_num].legend[i], buff );
	}

	/* don't bother drawing the bar if it's hidden */
        if (!XtIsManaged( bars[bar_num].box_widget))
	    return;

	/* update the bar */
        redraw_bar(bar_num);
    }

}

void set_foreground( w, color )
Widget w;
int color;
{
    switch (color) {
        case 0 : XSetForeground( XtDisplay(w), gc, green ); break;
        case 1 : XSetForeground( XtDisplay(w), gc, yellow ); break;
        case 2 : XSetForeground( XtDisplay(w), gc, red ); break;
/*
** MKN: Swapping magenta and blue
*/
        case 3 : XSetForeground( XtDisplay(w), gc, magenta ); break;
        case 4 : XSetForeground( XtDisplay(w), gc, blue ); break;
    }
}

void draw_proportion( bar_num )
int bar_num;
{
    register int i;
    int x, y;
    unsigned int length, height;
    long max_value,llength;  /* long is needed  for 128GB memory systems */
    Widget w = bars[bar_num].bar_widget;
    max_value = 0;

    for (i = 0; i < bars[bar_num].num_values; i++)
	max_value += bars[bar_num].values[i];

    if (max_value <=0) {
	/* No data -- clear widget  and return */
	XClearWindow(XtDisplay(w), XtWindow(w));
	return;
    }

    height = *bar_height;
    x = 0;
    y = 0;
    for (i = 0; i < bars[bar_num].num_values; i++) {

	llength = ((long)*bar_width * (long)bars[bar_num].values[i] + max_value / 2) 
		/ max_value;
	length = (unsigned int)llength;

	set_foreground( w, i );
	XFillRectangle( XtDisplay(w), XtWindow(w), gc,
		x, y, length, height );

	x += length;
    }
}

void draw_proportion_limit( bar_num )
int bar_num;
{
    register int i;
    int x, y;
    unsigned int length, height;
    int max_value;
    Widget w = bars[bar_num].bar_widget;
    max_value = 0;

    for (i = 1; i < bars[bar_num].num_values-1; i++)
	max_value += bars[bar_num].values[i];

    if (max_value <=0) {
	/* No data -- clear widget  and return */
	XClearWindow(XtDisplay(w), XtWindow(w));
	return;
    }

    height = *bar_height;
    x = 0;
    y = 0;
    /* Draw bars except for limits */
    for (i = 1; i < bars[bar_num].num_values-1; i++) {

	length = (*bar_width * bars[bar_num].values[i] + max_value / 2) 
		/ max_value;

	set_foreground( w, i );
	XFillRectangle( XtDisplay(w), XtWindow(w), gc,
		x, y, length, height );

	x += length;
    }

    /* Draw Min */
    length = (*bar_width * bars[bar_num].values[0] + max_value / 2) 
		/ max_value;
    set_foreground( w, 0 );
    XFillRectangle( XtDisplay(w), XtWindow(w), gc,
	0, 0, length, height/2 );

    /* Draw Max */
    length = (*bar_width * bars[bar_num].values[bars[bar_num].num_values-1]
		 + max_value / 2) / max_value;
    x = *bar_width - length;
    set_foreground( w, bars[bar_num].num_values-1 );
    XFillRectangle( XtDisplay(w), XtWindow(w), gc,
	x, 0, length, height/2 );

}

void draw_multi_proportion( bar_num )
int bar_num;
{
    register int i, row;
    int x, y;
    unsigned int length, height;
    int max_value;
    Widget w = bars[bar_num].bar_widget;

/*
    XClearWindow(XtDisplay(w), XtWindow(w));
*/

    height = *bar_height-1;
    y = 0;
    for (row = 0; row < bars[bar_num].num_scales; row++) {

        x = 0;

	max_value = 0;
	for (i = 0; i < 5; i++)
	    max_value += bars[bar_num].multi_values[row][i];

	if (max_value <= 0) {
            /* Draw background bar */
	    length = *bar_width;

            XSetForeground( XtDisplay(w), gc, black );
	    XFillRectangle( XtDisplay(w), XtWindow(w), gc,
		            x, y, length, height );
	    
        }
        else {
          for (i = 0; i < 5; i++) {

	    length = (*bar_width * bars[bar_num].multi_values[row][i]
		+ max_value / 2) / max_value;

	    set_foreground( w, i );
	    XFillRectangle( XtDisplay(w), XtWindow(w), gc,
		x, y, length, height );

	    x += length;
          }
	}
	y += height+1;
    }
}

void draw_length( bar_num )
int bar_num;
{
    register int i;
    int x, y;
    unsigned int length, height;
    int max_value;
    int scale;
    Widget w = bars[bar_num].bar_widget;
    max_value = 0;

    XClearWindow(XtDisplay(w), XtWindow(w));

    for (i = 0; i < bars[bar_num].num_values; i++)
	max_value = MAX(bars[bar_num].values[i], max_value );

    if (max_value <=0) {
	/* No data, return */
	return;
    }

    scale = 100 * (max_value / 100);
    if (max_value % 100) scale += 100;

    height = *bar_height * bars[bar_num].num_values;

    x = 0;
    y = 0;
    for (i = 0; i < bars[bar_num].num_values; i++) {

	length = *bar_width * bars[bar_num].values[i]/scale;

	set_foreground( w, i );
	XFillRectangle( XtDisplay(w), XtWindow(w), gc,
		x, y, length, *bar_height );

	y += *bar_height;
    }

    /* add scale marks */

    XSetForeground( XtDisplay(w), gc, white );
    x = 0;
    y = 0;
/*
** MKN: Limit number of scale marks to 20
*/
    if (scale <= 2000) for (i = 100; i < scale; i += 100) {
	x = *bar_width  * i / scale;
	XDrawLine( XtDisplay(w), XtWindow( w ), gc,
		x, y, x, height);
    }
}

void redraw_bar( bar_num )
int bar_num;
{
    /* Draw appropriately color rectangles */

    if (!XtIsManaged( bars[bar_num].bar_widget ) )
	return;

    switch (bars[bar_num].bar_type) {
	case PROPORTION :
	    draw_proportion( bar_num );
	    break;
	case LENGTH :
	    draw_length( bar_num );
	    break;
	case MULTI_PROPORTION :
	    draw_multi_proportion( bar_num );
	    break;
	case PROPORTION_LIMIT :
	    draw_proportion_limit( bar_num );
	    break;
	default :
	    fprintf(stderr, "Bad type for bar %d in routine redraw_bar()\n", bar_num );
	    break;
    }
}

static XtTimerCallbackProc update_display();

display_bars( nbars )
int nbars;
{
    int i;

    for (i = 0; i < nbars; i++)
	    (*bars[i].update)();
}

do_widget_loop()
{
        timer = XtAppAddTimeOut( app_context, 
		(unsigned long) (interval * 1000), update_display,
		(caddr_t) NULL);
        XtAppMainLoop( app_context );
}

/* ARGSUSED */
static XtTimerCallbackProc
update_display(closure, id)
caddr_t closure;
XtIntervalId *id;
{
    display_bars(nbars);
    
    if (count > 0) {
	if (--count == 0) {
	    set_suspend_button( TRUE );
	    return;
	}
    }
    timer = XtAppAddTimeOut( app_context, 
		(unsigned long) (interval * 1000), update_display,
		(caddr_t) NULL);
}

start_timer()
{
    timer = XtAppAddTimeOut( app_context,
		(unsigned long) (interval * 1000), update_display,
		(caddr_t) NULL);
}
